part of katana_module;

/// Module class for generating Document query.
///
/// It is possible to override [build()] to generate a model.
@immutable
class DocumentQueryModule extends QueryModule<DynamicDocumentModel> {
  const DocumentQueryModule({required this.path});

  final String path;

  @override
  DynamicDocumentModel build(BuildContext context) {
    final path = context.convert(this.path);

    return context.model!.loadDocument(
      useProvider(context.model!.documentProvider(path)),
    );
  }

  @override
  DynamicMap toMap() {
    return {
      "type": "document",
      if (path.isNotEmpty) "path": path,
    };
  }

  static DocumentQueryModule? _from(DynamicMap map) {
    if (!map.containsKey("path")) {
      return null;
    }
    return DocumentQueryModule(
      path: map.get<String>("path", ""),
    );
  }
}

/// Module class for generating Collection query.
///
/// It is possible to override [build()] to generate a model.
@immutable
class CollectionQueryModule extends QueryModule<DynamicCollectionModel> {
  // ignore: unused_element
  const CollectionQueryModule._({this.path, this.query});
  const CollectionQueryModule.path(String path)
      // ignore: prefer_initializing_formals
      : path = path,
        query = null;
  const CollectionQueryModule.query(CollectionQuery query)
      : path = null,
        // ignore: prefer_initializing_formals
        query = query;

  final String? path;
  final CollectionQuery? query;

  @override
  @protected
  DynamicCollectionModel build(BuildContext context) {
    final path = context.convert(this.path ?? query!.value);

    return context.model!.loadCollection(
      useProvider(context.model!.collectionProvider(path)),
    );
  }

  @override
  DynamicMap toMap() {
    return {
      "type": "collection",
      if (path.isNotEmpty) "path": path,
      if (query != null) "query": query!.toMap(),
    };
  }

  static CollectionQueryModule? _from(DynamicMap map) {
    if (!map.containsKey("path") && !map.containsKey("query")) {
      return null;
    }
    return CollectionQueryModule._(
      path: map.get<String?>("path", null),
      query: map.getAsMap("query").toCollectionQuery(),
    );
  }
}

/// Module class for generating Query.
///
/// It is possible to override [build()] to generate a model.
@immutable
abstract class QueryModule<T extends Model> {
  const QueryModule();

  /// Build the model.
  ///
  /// Please inherit and use it.
  @protected
  T build(BuildContext context);

  /// Create a new model.
  ///
  /// Pass the BuildContext when you create it.
  ///
  /// Also, if you use hooks, please make sure that the hooks are loaded at the right time.
  T create(BuildContext context) => build(context);

  /// Convert the query to [DynamicMap].
  DynamicMap toMap();
}

extension _QueryModuleBuildContextExtensions on BuildContext {
  static final _converter = RegExp(r"\{([^\{\}]+?)\}");
  String convert(String path) {
    while (path.contains("{") || path.contains("}")) {
      path = path.replaceAllMapped(
        _converter,
        (match) {
          final command = match.group(1);
          if (command.isEmpty || !command!.contains(":")) {
            return "";
          }
          final split = command.split(":");
          final key = split.first;
          final value = split.last;
          switch (key) {
            case "context":
              return this[value]?.toString() ?? "";
            case "account":
              switch (value) {
                case "id":
                  return model?.userId ?? "";
              }
              return "";
          }
          return "";
        },
      );
    }
    return path;
  }
}

extension QueryModuleExtensions on DynamicMap? {
  /// Convert the query from [DynamicMap].
  QueryModule? toQueryModule() {
    if (isEmpty || !containsKey("type")) {
      return null;
    }

    switch (this!["type"] as String) {
      case "document":
        return DocumentQueryModule._from(this!);
      case "collection":
        return CollectionQueryModule._from(this!);
    }
    return null;
  }
}
